type interval = { tag: Tag.t; start: int; stop: int }
type intervals = interval list

type t = {
    colors: intervals;
    sizes: intervals;
    underlines1: intervals;
    underlines2: intervals;
    underlines3: intervals;
    emphasiss: intervals;
    italics: intervals;
    bolds: intervals;
    typewriters: intervals
  } 

let initial_intervals = {
    colors = [];
    sizes = [];
    underlines1 = [];
    underlines2 = [];
    underlines3 = [];
    emphasiss = [];
    italics = [];
    bolds = [];
    typewriters = []
  }

type context = {
    color: char;
    size: int;
    underline: int;
    emphasis: int;
    italic: bool;
    bold: bool;
    typewriter: bool
  } 

let initial_context = {
  color = '_';
  size = -1;
  underline = 0;
  emphasis = 0;
  italic = false;
  bold = false;
  typewriter = false
} 

let of_trees trees =
  let rec build context intervals tree =
    match tree with
      Tree.Node (tag, children) ->
	let child_context =
	  match tag with
	    Tag.B -> { context with bold = true }
	  | Tag.I -> { context with italic = true }
	  | Tag.PL -> { context with underline = 0; emphasis = 0; italic = false;
			bold = false; typewriter = false }
	  | Tag.TT -> { context with typewriter = true }
	  | Tag.U -> { context with underline = context.underline + 1 }
	  | Tag.EM x -> { context with emphasis = x }
	  | Tag.SIZE x -> { context with size = x }
	  | Tag.COLOR x -> { context with color = x }
	in List.fold_left (build child_context) intervals children
    | Tree.Space (start, stop) ->
	{ intervals with
	  colors =
	  if context.color = '_' then
	    if (context.underline = 0) then
	      { tag = Tag.PL; start = start; stop = stop }::intervals.colors
	    else
	      intervals.colors
	  else
	    { tag = if (context.underline = 0) then Tag.PL else Tag.COLOR context.color; start = start; stop = stop }::intervals.colors;
	  sizes =
	  if context.size = -1 then
	    intervals.sizes
	  else
	    { tag = Tag.SIZE context.size; start = start; stop = stop }::intervals.sizes;
	  underlines1 =
	  if context.underline < 1 then
	    intervals.underlines1
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines1;
	  underlines2 =
	  if context.underline < 2 then
	    intervals.underlines2
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines2;
	  underlines3 =
	  if context.underline < 3 then
	    intervals.underlines3
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines3;
	  typewriters =
	  if not context.typewriter then
	    intervals.typewriters
	  else
	    { tag = Tag.TT; start = start; stop = stop }::intervals.typewriters }
    | Tree.Chars (start, stop) ->
	{ colors =
	  if context.color = '_' then
	    intervals.colors
	  else
	    { tag = Tag.COLOR context.color; start = start; stop = stop }::intervals.colors;
	  sizes =
	  if context.size = -1 then
	    intervals.sizes
	  else
	    { tag = Tag.SIZE context.size; start = start; stop = stop }::intervals.sizes;
	  underlines1 =
	  if context.underline < 1 then
	    intervals.underlines1
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines1;
	  underlines2 =
	  if context.underline < 2 then
	    intervals.underlines2
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines2;
	  underlines3 =
	  if context.underline < 3 then
	    intervals.underlines3
	  else
	    { tag = Tag.U; start = start; stop = stop }::intervals.underlines3;
	  emphasiss =
	  if context.emphasis = 0 then
	    intervals.emphasiss
	  else
	    { tag = Tag.EM (context.emphasis); start = start; stop = stop }::intervals.emphasiss;
	  italics =
	  if not context.italic then
	    intervals.italics
	  else
	    { tag = Tag.I; start = start; stop = stop }::intervals.italics;
	  bolds =
	  if not context.bold then
	    intervals.bolds
	  else
	    { tag = Tag.B; start = start; stop = stop }::intervals.bolds;
	  typewriters =
	  if not context.typewriter then
	    intervals.typewriters
	  else
	    { tag = Tag.TT; start = start; stop = stop }::intervals.typewriters } in
  let rec concat_equal = function
      {tag = t1; start = s1; stop = e1}::{tag = t2; start = s2; stop = e2}::tl
      when (e2 = s1 && t1 = t2) ->
	concat_equal ({ tag = t1; start = s2; stop = e1 }::tl)
    | hd::tl -> hd::(concat_equal tl)
    | [] -> [] in
  let raw_intervals = List.fold_left (build initial_context) initial_intervals trees in
  { colors = List.rev (concat_equal raw_intervals.colors);
    sizes = List.rev (concat_equal raw_intervals.sizes);
    underlines1 = List.rev (concat_equal raw_intervals.underlines1);
    underlines2 = List.rev (concat_equal raw_intervals.underlines2);
    underlines3 = List.rev (concat_equal raw_intervals.underlines3);
    emphasiss = List.rev (concat_equal raw_intervals.emphasiss);
    italics = List.rev (concat_equal raw_intervals.italics);
    bolds = List.rev (concat_equal raw_intervals.bolds);
    typewriters = List.rev (concat_equal raw_intervals.typewriters)
  }

let print_interval fmt { tag = tag; start = start; stop = stop } =
  begin
    Format.fprintf fmt "(";
    Tag.print fmt tag;
    Format.fprintf fmt " %d-%d)" start stop
  end

let rec print_intervals fmt intervals =
  begin
    Format.fprintf fmt "@[";
    List.iter
      (fun interval -> print_interval fmt interval; Format.fprintf fmt "@ ")
      intervals;
    Format.fprintf fmt "@]"
  end

let rec print fmt canon = 
  begin
    Format.pp_print_string fmt "colors = ";
    print_intervals fmt canon.colors;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "sizes = ";
    print_intervals fmt canon.sizes;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "underlines1 = ";
    print_intervals fmt canon.underlines1;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "underlines2 = ";
    print_intervals fmt canon.underlines2;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "underlines3 = ";
    print_intervals fmt canon.underlines3;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "emphasiss = ";
    print_intervals fmt canon.emphasiss;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "italics = ";
    print_intervals fmt canon.italics;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "bolds = ";
    print_intervals fmt canon.bolds;
    Format.pp_force_newline fmt ();
    Format.pp_print_string fmt "typewriters = ";
    print_intervals fmt canon.typewriters;
    Format.pp_force_newline fmt ()
  end
