
(* split1 < split 2 *)
(* split1 splits node 1 *)
type edge = { mutable n1: node; mutable split1: int; mutable n2: node; mutable split2: int }
and node = { interval: Canonical.interval; mutable edges: edge list; idx: int }

type leave_or_interval =
    LEAVE of (Tree.t * int * int)
  | INTERVAL of (Canonical.interval * int * int)

open Canonical

module OrderedNode =
  struct
    type t = node
    (* 1 are small (good, correct)
       0 are big (correct)
       short intervalls are small (correctness)
       high length is small (heuristic)
       differentiate at all costs
    *)
    let compare n1 n2 =
      match (List.length n1.edges, List.length n2.edges) with
	1, 1 -> 
	  let a = compare n1.interval n2.interval in
	  if a !=0 then a else (n1.idx - n2.idx)
      |	1, n -> -1
      | n, 1 -> 1
      |	0, 0 -> 
	  let a = compare n1.interval n2.interval in
	  if a !=0 then a else (n1.idx - n2.idx)
      |	0, n -> 1
      | n, 0 -> -1
      |	n, m -> 
	  let b = n1.interval.stop - n1.interval.start 
	      - n2.interval.stop + n2.interval.start in
	  if b != 0 then 
	    b 
	  else if n != m then
	    m - n
	  else
	    let a = compare n1.interval n2.interval in
	    if a !=0 then a else (n1.idx - n2.idx)
  end

module NodeSet = Set.Make (OrderedNode)

(*
let info_list trees =
  let rec rinfo_list trees acc =
    match trees with
      [] -> acc
    | hd::tl ->
	begin
	  match hd with 
	    Tree.Node (tag, children) ->
	      rinfo_list tl (rinfo_list children acc)
	  | x -> rinfo_list tl (x::acc)
	end
  in List.rev (rinfo_list trees [])
*)
let of_canonical flag old_trees canon text_length =

  let print_nodes fmt nodes =
    begin
      Format.fprintf fmt "@[";
      NodeSet.iter
	(fun node -> Canonical.print_interval fmt node.interval; Format.fprintf fmt "@ ")
	nodes;
      Format.fprintf fmt "@]"
    end in

  (* compute dont cares from old tree *)
  let dont_cares = 
    let rec rinfo_list trees acc =
      match trees with
	[] -> acc
      | hd::tl ->
	  begin
	    match hd with 
	      Tree.Node (tag, children) ->
		rinfo_list tl (rinfo_list children acc)
	    | x -> rinfo_list tl (x::acc)
	  end
    in List.rev (rinfo_list old_trees []) in
  let dc = List.rev (List.fold_left (fun r e -> match e with
    Tree.Space (start, stop) -> {tag = Tag.PL; start = start; stop = stop}::r
  | Tree.Chars (start, stop) -> r
  | _ -> failwith "no proper dont care") [] dont_cares)
  in let cdc =
    let rec rcdc list = match list with 
      [] -> []
    | a::b::tl when a.stop = b.start ->
	rcdc ({tag = Tag.PL; start = a.start; stop = b.stop }::tl)
    | a::tl ->
	a::(rcdc tl)
    in rcdc dc
  in
  (*
  let rec combine list = 
    match list with
      a::b::tl when a.stop != b.start ->
	a:: (combine (b::tl))
    | a::b::c::tl when a.tag = c.tag && b.stop = c.start ->
	b:: (combine ({tag = a.tag; start = a.start; stop = c.stop }::tl))
    | a::tl -> a::(combine tl)
    | [] -> []
  in
  *)
  let sort_list list = List.sort (fun x y -> if x.start = y.start then y.stop - x.stop else x.start - y.start) list
  in

  let combine_pl list =
    let input = sort_list (cdc @ list) in
    let rec rcombine_pl list = 
      match list with
	a::b::c::tl when a.stop = b.start && b.stop = c.start && b.tag = Tag.PL && a.tag = c.tag ->
	  (rcombine_pl ({tag = a.tag; start = a.start; stop = c.stop }::tl))
      |	a::tl when a.tag = Tag.PL ->
	  (rcombine_pl tl)
      |	a::tl -> a::(rcombine_pl tl)
      | [] -> []
    in rcombine_pl input
  in
  let combine_c list =
    let input = list in
    let rec rcombine_c list = 
      match list with
	a::b::tl when a.stop = b.start && b.tag = Tag.PL ->
	  (rcombine_c ({tag = a.tag; start = a.start; stop = b.stop }::tl))
      |	a::b::tl when a.stop = b.start && a.tag = Tag.PL ->
	  (rcombine_c ({tag = b.tag; start = a.start; stop = b.stop }::tl))
      |	a::tl when a.tag = Tag.PL ->
	  (rcombine_c tl)
      |	a::tl -> a::(rcombine_c tl)
      | [] -> []
    in rcombine_c input
  in
(*
  print_intervals Format.std_formatter (sort_list (combine (canon.colors)));
  *)  
  (* preprocess canonical form *)
  let lists = 
    if flag then [| combine_c canon.colors; canon.sizes; canon.underlines1;
		 canon.underlines2; canon.underlines3; combine_pl canon.emphasiss; 
		 combine_pl canon.italics; combine_pl canon.bolds; 
		 canon.typewriters; |] 
	else [| combine_c canon.colors; canon.sizes; canon.underlines1;
		 canon.underlines2; canon.underlines3; canon.emphasiss; 
		 canon.italics; canon.bolds; 
		 canon.typewriters; |]
  in

  (* construct graph *)
  let nodes = 
    let rec loop node_table =
      (* finds head with lowest start *)
      let min_interval_idx () =
	let rec search min_idxc idx =
	  if idx >= Array.length lists then
	    min_idxc
	  else match lists.(idx) with
	    [] -> search min_idxc (succ idx)
	  | hd::tl ->
	      if min_idxc < 0 || hd.start < (List.hd lists.(min_idxc)).start then
		search idx (succ idx)
	      else
		search min_idxc (succ idx)
	in search (-1) 0 in
      (* compute minimal index *)
      let min_idx = min_interval_idx () in
      if min_idx < 0 then (* done *)
	node_table
      else
	begin
	  (* match minimal node *)
	  match lists.(min_idx) with
	    hd::tl ->
	      begin
		let get_node interval idx = 
		  if Hashtbl.mem node_table (interval, idx) then
		    Hashtbl.find node_table (interval, idx)
		  else (
		    let new_node = { interval = interval; edges = []; idx = idx } in
		    Hashtbl.add node_table (interval, idx) new_node;
		    new_node
		   ) in
		let nnode = get_node hd min_idx in
	        (* compute edges that overlap hd from idx on *)
		let rec edges idx =
		  if idx >= Array.length lists then
		    (* [] *) ()
		  else
		    begin
		      (* print_string "hello"; *)
		      (* construct all new edges that interfere with hd in list *)
		      let rec edges1 list idx =
			match list with
			  [] -> () (* [] *)
			| hd1::tl1 ->
			    if hd1.start < hd.stop then
			      begin
			         (*
				 print_newline ();
				 print_string "Xxx: ";
				 print_int hd1.start;
				 print_string " ";
				 print_int hd1.stop;
				 print_string " ";
				 print_int hd.start;
				 print_string " ";
				 print_int hd.stop;
				 print_string " ";
			         *)
			        (* the list remains interesting *)
				if hd1.start > hd.start 
				    &&  hd1.stop > hd.stop then
				  begin
			            (* overlap! get other node *)
				    let onode = get_node hd1 idx in
			            (* construct edge *)
				    let nedge = ({ n1 = nnode; split1 = hd1.start; n2 = onode; split2 = hd.stop}) in
				    (*
				     print_string "new edge "; 
				     print_int hd1.start; 
				     print_string " "; 
				     print_int hd.stop;
				     print_newline();
				   *)
				    nnode.edges <- nedge::nnode.edges;
				    onode.edges <- nedge::onode.edges;
				    (* nedge::(edges1 tl1 idx) *)
				    (edges1 tl1 idx)
				  end
				else
				  (edges1 tl1 idx)
			      end
			    else
			      (* we have seen enough of this list *)
			      () (*[]*)
		      in (edges1 lists.(idx) idx);(edges (succ idx))
		    end
		in
		edges 0; 
		lists.(min_idx) <- tl;
		loop node_table
	      end
	  | _ -> failwith "Empty minimum list";
	end
    in loop (Hashtbl.create 100000) in
  (* node set *)
  let node_set = ref (NodeSet.empty) in
  Hashtbl.iter (fun iv node ->
    node_set := NodeSet.add node !node_set;
	       ) nodes;
  let rec split_nodes () =
    let min_node = NodeSet.min_elt !node_set in
    if List.length (min_node.edges) = 0 then
      begin
	(*
	NodeSet.iter (fun node ->
	  if ((List.length node.edges) !=0) then
	    failwith "egdgecount != 0"
		     ) !node_set;
	   *)
	()
      end
    else
      let (split_node, split_edge) = 
	if (List.length (min_node.edges)) = 1 then
	  if (List.hd (min_node.edges)).n1 == min_node then
	    ((List.hd (min_node.edges)).n2, (List.hd (min_node.edges)))
	  else
	    ((List.hd (min_node.edges)).n1, (List.hd (min_node.edges)))
	else
	  (min_node,(List.hd (min_node.edges)))  in
      begin
        (* split node split_node *)
	node_set := NodeSet.remove split_node !node_set;
	(*
	print_string "split ";
	print_int split_node.interval.start;
	print_string " ";
	print_int split_node.interval.stop;
	print_string " ";
	print_int (List.length split_node.edges);
	print_newline();
	print_int min_node.interval.start;
	print_string " ";
	print_int min_node.interval.stop;
	print_string " ";
	print_int (List.length min_node.edges);
	print_newline();
	*)
	(* find a good split point, here take first edge *)
	let split_point = 
	  if split_edge.n1 == split_node then
	    split_edge.split1
	  else
	    split_edge.split2 in
	(*
	print_string "splitpoint ";
	print_int split_point;
	print_newline();
	   *)
	let new_left_node = { 
	  interval = { tag = split_node.interval.tag; start = split_node.interval.start; stop = split_point };
	  edges = [];
	  idx = split_node.idx
	} in
	let new_right_node = { 
	  interval = { tag = split_node.interval.tag; start = split_point; stop = split_node.interval.stop };
	  edges = [];
	  idx = split_node.idx
	} in
	let rec treat_edges edges =
	  match edges with
	    [] -> ()
	  | hd ::tl ->
	      if (hd.n1 == split_node) then
		begin
		  node_set := NodeSet.remove hd.n2 !node_set;
		  if (split_point = hd.split1) then
		    (* remove this edge *)
		    (hd.n2).edges <- List.filter (fun x -> x != hd) (hd.n2).edges
		  else if (split_point < hd.split1) then
		    (* edge now with right *)
		    begin
		      hd.n1 <- new_right_node;
		      (hd.n1).edges <- hd::(hd.n1).edges
		    end
		  else
		    (* edge now with left *)
		    begin
		      hd.n1 <- new_left_node;
		      hd.split2 <- split_point;
		      (hd.n1).edges <- hd::(hd.n1).edges
		    end;
		  node_set := NodeSet.add hd.n2 !node_set;
		end
	      else
		begin
		  node_set := NodeSet.remove hd.n1 !node_set;
		  if (split_point = hd.split2) then
		    (* remove this edge *)
		    (hd.n1).edges <- List.filter (fun x -> x != hd) (hd.n1).edges
		  else if (split_point < hd.split2) then
		    (* edge now with right *)
		    begin
		      hd.n2 <- new_right_node; 
		      hd.split1 <- split_point;
		      (hd.n2).edges <- hd::(hd.n2).edges
		    end
		  else
		    (* edge now with left *)
		    begin
		      hd.n2 <- new_left_node;
		      (hd.n2).edges <- hd::(hd.n2).edges
		    end;
		  node_set := NodeSet.add hd.n1 !node_set;
		end;
	      treat_edges tl;
	in treat_edges split_node.edges;
	node_set := NodeSet.add new_left_node !node_set;
	node_set := NodeSet.add new_right_node !node_set;
	split_nodes();
      end
  in  
  split_nodes();
  (* construct a list with the new intervals *)
  let new_intervals = NodeSet.fold (fun nd l -> nd.interval::l) !node_set [] in
  let sorted_intervals = List.sort (fun i1 i2 -> 
    if (i1.start = i2.start) then 
      i2.stop - i1.stop
    else
      i1.start - i2.start) new_intervals in

  (* build tree *)
  let lf_start nd = match nd with
    Tree.Space (start, stop) -> start
  | Tree.Chars (start, stop) -> start
  | _ -> failwith "a node" in
  let lf_stop nd = match nd with
    Tree.Space (start, stop) -> stop
  | Tree.Chars (start, stop) -> stop
  | _ -> failwith "a node" in
  let m_start nd = match nd with
    LEAVE (_, start, stop) -> start
  | INTERVAL (_, start, stop) -> start in
  let m_stop nd = match nd with
    LEAVE (_, start, stop) -> stop
  | INTERVAL (_, start, stop) -> stop in
  (* construct list with intervals and dont cares *)
  let merged = 
    let rec merge_them iv lv =
      match iv with
	[] -> 
	  begin
	    match lv with
	      [] -> []
	    | hdl::tll ->
		(LEAVE (hdl, lf_start hdl, lf_stop hdl)) :: (merge_them [] tll) 
	  end
      | hdi::tli ->
	  begin
	    match lv with
	      [] ->
		(INTERVAL (hdi, hdi.start, hdi.stop)) :: (merge_them tli []) 
	    | hdl::tll ->
		if (lf_start hdl < hdi.start) then
		  (LEAVE (hdl, lf_start hdl, lf_stop hdl)) :: (merge_them iv tll) 
		else
		  (INTERVAL (hdi, hdi.start, hdi.stop)) :: (merge_them tli lv) 
	  end
    in merge_them sorted_intervals dont_cares in
  (* construct tree *)
  let rec mk_tree trees all stop =
    match all with
      [] -> (trees, all)
    | hd::tl ->
	if (m_start hd) >= stop then 
	  (trees, all)
	else
	  begin
	    match hd with
	      LEAVE (lf, _, _) -> 
		mk_tree (lf::trees) tl stop
	    | INTERVAL (iv, _, _) -> 
		let (ntrees, nintervals) = mk_tree [] tl iv.stop in
		mk_tree ((Tree.Node (iv.tag, (List.rev ntrees)))::trees) nintervals stop
	  end
  in 
  (*
  List.map (fun l -> match l with
    LEAVE (_, start, stop) -> 
      Format.pp_print_string Format.std_formatter "(LEAVE ";
      Format.pp_print_int Format.std_formatter start;
      Format.pp_print_string Format.std_formatter " ";
      Format.pp_print_int Format.std_formatter stop;
      Format.pp_print_string Format.std_formatter ")"
  | INTERVAL (iv, start, stop) -> Canonical.print_interval Format.std_formatter iv
	   ) merged;
     *)
(*
  Format.pp_print_string Format.std_formatter "INPUT";
  Canonical.print Format.std_formatter canon; 
  Canonical.print_intervals (Format.std_formatter) sorted_intervals;
  Format.pp_force_newline (Format.std_formatter) ();*)
  List.rev (fst (mk_tree [] merged text_length))

(*
let _ = 
  let (text,trees) = Parser.parse (Stream.of_channel stdin) in
  let opt_canonical = Canonical.of_trees trees in
  let graph_trees = of_canonical true trees opt_canonical (String.length text) in
  let graph_canonical = Canonical.of_trees graph_trees in
(*
  Format.pp_print_string Format.std_formatter "OUTPUT";
  Canonical.print Format.std_formatter graph_canonical;
*)
  Printtree.print_tree_with_stack graph_trees text;

  prerr_string "After parsing: ";
  prerr_int (Tree.sizes trees);
  prerr_newline ();
  prerr_string "After graph: ";
  prerr_int (Tree.sizes graph_trees);
  prerr_newline ();
  if (graph_canonical = opt_canonical) then
    begin
      prerr_string "Equal!";
      prerr_newline ()
    end
  else
    begin
      prerr_string "False!";
      prerr_newline ()
    end

;;
*)
