let read_file stream =
  let buffer = Buffer.create 10000 in
  begin
    try
      while true do
        Buffer.add_char buffer (Stream.next stream)
      done
    with
      Stream.Failure -> ()
  end;
  Buffer.contents buffer

let get_total_time () =
  let args = Sys.argv in
  if Array.length args > 1 then int_of_string (args.(1)) else 180

let read_text = ref None
let best_tree = ref None

let print_and_exit () =
  begin
    match (!best_tree) with
    | None ->
        begin
          match (!read_text) with
          |  None ->
              ()
          |  Some source ->
              print_string source
        end
    |  Some (size, trees, canonical, text) ->
        begin
          match (!read_text) with
          |  None ->
              Printtree.print_tree_with_stack trees text
          |  Some source ->
              if (size + String.length text < String.length source) then
                Printtree.print_tree_with_stack trees text
              else
                print_string source
        end
  end;
  exit 0

let handle_signal _ =
  print_and_exit ()

type compare_result =
    Unchecked of Tree.t list * int
  | Better of Tree.t list * Canonical.t * int
  | Incorrect

let install_new_trees new_trees =
  match !best_tree with
  | None ->
      print_and_exit ()
  | Some (old_size, old_trees, old_canonical, text) ->
      let new_size = Tree.sizes new_trees in
      if (new_size > old_size) then
	begin
	  output_string stderr "bigger";flush stderr;
          Unchecked (new_trees, new_size)
	end
      else
        let new_canonical = Canonical.of_trees new_trees in
        if new_canonical <> old_canonical then
	  begin
	    output_string stderr "incorrect";flush stderr; 
            Incorrect
	  end
        else
          begin
            best_tree := Some (new_size, new_trees, new_canonical, text);
            Better (new_trees, new_canonical, new_size)
          end

(* Applies the best tree to the optimisation function *)
let take_best f (best_trees, best_canonical, best_size) last =
  f best_trees best_canonical best_size

(* Applies the last tree to the optimisation function *)
let take_last f (best_trees, best_canonical, best_size) last =
  match last with
    Unchecked (last_trees, last_size) ->
      let last_canonical = Canonical.of_trees last_trees in
      if (last_canonical = best_canonical) then
        f last_trees last_canonical last_size
      else
        f best_trees best_canonical best_size
  | Better (last_trees, last_canonical, last_size) ->
      f last_trees last_canonical last_size
  | Incorrect ->
      f best_trees best_canonical best_size


(* Applies the list of optimisation functions *)
(* The head function is applied first *)
let apply_optimisations trees canonical size =
  let rec apply_one best last = function
      [] ->
        ()
    | f::fs ->
        let new_trees = f best last in
        let new_last = install_new_trees new_trees in
        let (best_size, best_trees, best_canonical, _) = 
          match !best_tree with
            None -> print_and_exit ()
          | Some b -> b in
        let new_best = (best_trees, best_canonical, best_size) in
        apply_one new_best new_last fs 
  in
  apply_one (trees, canonical, size) Incorrect

let main () =
  (* Read the input *)
  let start_date = Unix.time () in
  let source_text = read_file (Stream.of_channel stdin) in
  let read_date = Unix.time () in
  read_text := Some source_text;

  begin
    try
      (* Install signal handlers *)
      Sys.set_signal Sys.sigabrt (Sys.Signal_handle handle_signal);
      Sys.set_signal Sys.sigalrm (Sys.Signal_handle handle_signal);
      Sys.set_signal Sys.sigfpe (Sys.Signal_handle handle_signal);
      (*Sys.set_signal Sys.sighup (Sys.Signal_handle handle_signal);*)
      Sys.set_signal Sys.sigill (Sys.Signal_handle handle_signal);
      (*Sys.set_signal Sys.sigint (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigkill (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigpipe (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigquit (Sys.Signal_handle handle_signal);*)
      Sys.set_signal Sys.sigsegv (Sys.Signal_handle handle_signal);
      (*Sys.set_signal Sys.sigterm (Sys.Signal_handle handle_signal);*)
      Sys.set_signal Sys.sigusr1 (Sys.Signal_handle handle_signal);
      Sys.set_signal Sys.sigusr2 (Sys.Signal_handle handle_signal);
      Sys.set_signal Sys.sigchld (Sys.Signal_handle handle_signal);
      (*Sys.set_signal Sys.sigcont (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigstop (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigtstp (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigttin (Sys.Signal_handle handle_signal);*)
      (*Sys.set_signal Sys.sigttou (Sys.Signal_handle handle_signal);*)
      Sys.set_signal Sys.sigvtalrm (Sys.Signal_handle handle_signal);
      (*Sys.set_signal Sys.sigprof (Sys.Signal_handle handle_signal);*)

      (* Install alarm *)
      let total_time = get_total_time () in
      (* !!! Here we assume that writing takes 10 times more time than *)
      (* !!! reading. *)
      let write_time = min (int_of_float ((read_date -. start_date) *. 10.) + 1) 10 in
      let optimize_time = total_time - write_time in
      if (optimize_time <= 0) then print_and_exit ();
      let _ = Unix.alarm optimize_time in

      (* Parse the input *)
      
      (* output_string stderr "parsing";flush stderr; *)
      let (parse_text,parse_trees)=Parser.parse(Stream.of_string source_text)in
	(* output_string stderr "parsed";flush stderr; *)
	let parse_canonical = Canonical.of_trees parse_trees in
	  (* output_string stderr "canonical";flush stderr; *)
	let parse_size = Tree.sizes parse_trees in
	  (* output_string stderr "size";flush stderr; *)
	best_tree := Some (parse_size, parse_trees, parse_canonical, parse_text);
      (* Apply the optimisations *)
      apply_optimisations parse_trees parse_canonical parse_size [
	take_best (Graph.of_canonical true);
        take_best Peep_hole_opt.remove_tags_trees;
        take_best Peep_hole_opt.collapse_trees (*;*)
	(* take_best Agent.optimize *)];

    with _ ->
      ();
  end;

  Sys.set_signal Sys.sigabrt (Sys.Signal_default);
  Sys.set_signal Sys.sigalrm (Sys.Signal_default);
  Sys.set_signal Sys.sigfpe (Sys.Signal_default);
(*Sys.set_signal Sys.sighup (Sys.Signal_default);*)
  Sys.set_signal Sys.sigill (Sys.Signal_default);
(*Sys.set_signal Sys.sigint (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigkill (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigpipe (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigquit (Sys.Signal_default);*)
  Sys.set_signal Sys.sigsegv (Sys.Signal_default);
(*Sys.set_signal Sys.sigterm (Sys.Signal_default);*)
  Sys.set_signal Sys.sigusr1 (Sys.Signal_default);
  Sys.set_signal Sys.sigusr2 (Sys.Signal_default);
  Sys.set_signal Sys.sigchld (Sys.Signal_default);
(*Sys.set_signal Sys.sigcont (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigstop (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigtstp (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigttin (Sys.Signal_default);*)
(*Sys.set_signal Sys.sigttou (Sys.Signal_default);*)
  Sys.set_signal Sys.sigvtalrm (Sys.Signal_default);
(*Sys.set_signal Sys.sigprof (Sys.Signal_default);*)

  print_and_exit ()
;;


let _ = 
(*  output_string stderr "start";
  flush stderr;*)
  main ();;
(*
  let (text,trees) = Parser.parse (Stream.of_channel stdin) in
  let canonical = Canonical.of_trees trees in
  let opt_trees = Peep_hole_opt.collapse_trees trees canonical 0 in
  let canon2 = Canonical.of_trees opt_trees in
    if canonical = canon2 then output_string stderr "correct"
    else output_string stderr "incorrect";
    Printtree.print_tree_with_stack opt_trees text
;;
*)




