(* local optimizer *)
type opt_tree =
(* the list of tags is sorted according to weight_tag  *)
(* there is at most one tag of each kind except for U *)
  | Node of ((Tag.t list) * (opt_tree list))
  | Space of (int * int)
  | Chars of (int * int)
  | Concat of opt_tree list (* it should only contain leaves *)
;;

let rec filter_option = function
  | [] -> []
  | x::xs ->
      match x with
	| None -> filter_option xs
	| Some y -> y::(filter_option xs)
;;

let rec rm_tags_tree = function
  | Tree.Node(tag,children) ->
      begin
	let childs = rm_tags_trees children in
	  match childs with
	    | [] -> None
	    | _ -> Some(Tree.Node(tag,childs))
      end
  | _ as t -> Some(t)
and rm_tags_trees trees = filter_option (List.map rm_tags_tree trees)
;;

let remove_tags_trees trees canonical size = rm_tags_trees trees;;

let weight_tag = function
  | Tag.PL -> 0
  | Tag.B -> 1
  | Tag.I -> 2
  | Tag.U -> 3
  | Tag.TT -> 4
  | Tag.EM _ -> 5
  | Tag.SIZE _ -> 6
  | Tag.COLOR _ -> 7
;;

let compare_tag t1 t2 = compare (weight_tag t1) (weight_tag t2);;

let insert_sorted_tag tag tags =
  let w = weight_tag tag in
  let rec insert_aux = function
    | [] -> [tag]
    | t::tags ->
	if (weight_tag t) < w then t::(insert_aux tags)
	else tag::t::tags
  in
    insert_aux tags
;;


let rec filter_and_remove_first p = function
  | [] -> raise Not_found
  | x::xs ->
      if p x then (x,xs)
      else 
	let (e,yss) = filter_and_remove_first p xs in
	  (e,x::yss)
;;


(* insert tag "at the beginning" *)
let insert_tag tag tags =
(* note that if PL appears in tags, it is in first position, always *)
  match tag with
    | Tag.B 
    | Tag.I 
    | Tag.PL
    | Tag.TT ->
	begin
	  match tags with
	    | Tag.PL::_ -> tags
	    | _ -> 
		if List.mem tag tags then tags
		else insert_sorted_tag tag tags
	end
    | Tag.U ->
	begin
	  match tags with
	    | Tag.PL::_ -> tags
	    | _ -> 
		let nb_u=List.fold_left (fun nb_u t -> if t = Tag.U then nb_u + 1 else nb_u) 0 tags in
		  if nb_u < 3 then insert_sorted_tag tag tags
		  else tags
	end
    | Tag.EM n ->
	begin
	  match tags with
	    | Tag.PL::_ -> tags
	    | _ ->
		begin
		  match n with
		    | 2 -> insert_sorted_tag tag (List.filter (function Tag.EM _ -> false | _ -> true) tags)
		    | _ -> 
			begin
			  try
			    begin
			      (* we remove the first and only EM from tags *)
			      (* S is shifted to the front *)
			      let (t,l) = filter_and_remove_first (function Tag.EM _ -> true | _ -> false) tags in
				match t with 
				  | Tag.EM 2 -> insert_sorted_tag t l
				  | _ -> l
			    end
			  with
			    | Not_found -> insert_sorted_tag tag tags
			end
		end
	end
    | Tag.SIZE n ->
	begin
	  if List.exists (function Tag.SIZE _ -> true | _ -> false) tags then
	    tags
	  else
	    insert_sorted_tag tag tags
	end
    | Tag.COLOR c ->
	begin
	  if List.exists (function Tag.COLOR _ -> true | _ -> false) tags then
	    tags
	  else
	    insert_sorted_tag tag tags
	end
;;

let encode_tag tag = [tag];;

(* this function should be called only after rm_tags_tree *)
let rec op_tree_of_tree = function
  | Tree.Node(tag,children) ->
      begin
	match children with
	  | [child] ->
	      begin
		let opt_child=op_tree_of_tree child in
		  match opt_child with
		    | Node(tags,childs) ->
			begin
			  Node(insert_tag tag tags,childs)
			end
		    | _ -> Node(encode_tag tag,[opt_child])
	      end
	  | _ -> Node(encode_tag tag,List.map op_tree_of_tree children)  
      end
  | Tree.Space(n,m) -> Space(n,m)
  | Tree.Chars(n,m) -> Chars(n,m)
;;

let rec flatten_map f = function
  | [] -> []
  | l::ls -> (f l)@(flatten_map f ls)
;;

let flatten_map f l = List.flatten (List.map f l);;

let rec tree_of_op_tree = function
  | Node(tags,children) ->
      let childs=flatten_map tree_of_op_tree children in
	List.fold_right (fun tag trees -> [Tree.Node(tag,trees)]) tags childs
  | Space(n,m) -> [Tree.Space(n,m)]
  | Chars(n,m) -> [Tree.Chars(n,m)]
  | Concat(l) -> flatten_map tree_of_op_tree l 
;;

let rec is_sublist xss yss =
  match xss with
  | [] -> true
  | x::xs -> (List.mem x yss) && (is_sublist xs yss)
;;

let rec insert_into_tags tags = function
  | [] -> tags
  | t::tail -> insert_sorted_tag t (insert_into_tags tags tail)
;;

(* take a tag and a list of tags *)
(* 
   return (less,greater,common) where 
   less are the tags (of weight) less than tag, 
   greater are the tags (of weight) greater or equal to tag (except one occurence of tag if exists),
   common indicates if tag appears in tags
*)
let discriminate tag tags = 
  let w = weight_tag tag in
  let rec discr_aux = function
    | [] -> ([],[],false)
    | t::tail -> 
	let (less,greater,common) = discr_aux tail in
	  if (weight_tag t) < w then
	    (t::less,greater,common)
	  else
	    if (weight_tag t) > w then
	      (less,t::greater,common)
	    else
	      if tag = t then
		if common then (less,t::greater,common)
		else (less,greater,true)
	      else
		(less,t::greater,common)
  in
    discr_aux tags
;;
      
let delta xss yss =
  (* note that U can appear several times, so we have to be careful of that *)
  (* xss INTER yss , xss \ yss, yss \ xss *)
  let (xss_u,xss_others) = List.partition (function Tag.U -> true | _ -> false) xss
  and (yss_u,yss_others) = List.partition (function Tag.U -> true | _ -> false) yss in
  let (inter,diff1)=List.partition (function t -> List.mem t yss_others) xss_others in
  let diff2=List.filter (function t -> not (List.mem t xss_others)) yss_others in
  let rec gen_u = function
    | 0 -> []
    | n -> Tag.U :: (gen_u (n-1))
  in
  let n=List.length xss_u and m=List.length yss_u in
  let inter_u=min n m in
    (insert_into_tags inter (gen_u inter_u),insert_into_tags diff1 (gen_u (n-inter_u)),insert_into_tags diff2 (gen_u (m-inter_u)))
;;

let color_and_size tags = 
  (* color and size, other tags *)
  List.partition (function Tag.SIZE _ | Tag.COLOR _ -> true | _ -> false) tags
;;

let rec join_trees = function
  | [] -> []
  | Node(tags,children)::tail ->
      begin
	let new_trees=join_trees tail in
	  match new_trees with
	    | Node(otags,ochildren)::rest ->
		let (inter_tags,diff_tags,diff_otags)=delta tags otags in
		  if (List.mem Tag.PL diff_tags) || (List.mem Tag.PL diff_otags) then 
		    (* here, PL tag is only in one of the two Node *)
		    (* we might only merge color and/or size *)
		    begin
		      let (col_size_tags,other_tags)=color_and_size tags 
		      and (col_size_otags,other_otags)=color_and_size otags in
		      let (inter_tags,diff_tags,diff_otags)=delta col_size_tags col_size_otags in
			begin
			  if inter_tags = [] then
			    begin
			      (* one of tags or otags is not empty, at least (because there is PL) *)
			      Node(tags,join_trees children)::new_trees
			    end
			  else
			    begin
			      let rest_tags = insert_into_tags other_tags diff_tags
			      and rest_otags = insert_into_tags other_otags diff_otags in
				(* at least, rest_tags or rest_otags is not empty (PL tag is here) *)
				begin
				  if rest_tags = [] then
				    begin
				      (* tags <= otags *)
				      Node(inter_tags,join_trees (children@[Node(rest_otags,ochildren)]))::rest
				    end
				  else
				    begin
				      if rest_otags = [] then
					begin
					  (* otags <= tags *)
					  Node(inter_tags,join_trees (Node(rest_tags,children)::ochildren))::rest
					end
				      else
					Node(inter_tags,[Node(rest_tags,join_trees children);Node(rest_otags,ochildren)])::rest
				    end
				end
			    end
			end
		    end
		  else
		    (* here PL tag is either in both or in none *) 
		    begin
		      if inter_tags = [] then
			begin
			  if (diff_tags = []) && (diff_otags = []) then
			    (join_trees (children@ochildren))@rest
			  else
			    Node(tags,join_trees children)::new_trees
			end
		      else
			begin
			  if (diff_tags = []) && (diff_otags = []) then
			    (* otags = tags *)
			    Node(tags,join_trees (children@ochildren))::rest
			  else
			    begin
			      if diff_tags = [] then
				begin
				  (* tags <= otags *)
				  Node(inter_tags,join_trees (children@[Node(diff_otags,ochildren)]))::rest
				end
			      else
				begin
				  if diff_otags = [] then
				    begin
				      (* otags <= tags *)
				      Node(inter_tags,join_trees (Node(diff_tags,children)::ochildren))::rest
				    end
				  else
				    Node(inter_tags,[Node(diff_tags,join_trees children);Node(diff_otags,ochildren)])::rest
				end
			    end
			end
		    end
	    | _ -> Node(tags,join_trees children)::new_trees
      end
  | Chars(n,m)::rest ->
      begin
	let new_trees = join_trees rest in
	  match new_trees with
	    | Chars(p,q)::tail -> Concat([Chars(n,m);Chars(p,q)])::tail 
	    | Space(p,q)::tail -> Concat([Chars(n,m);Space(p,q)])::tail
	    | Concat(l)::tail -> Concat(Chars(n,m)::l)::tail
	    | _ -> Chars(n,m)::new_trees
      end
  | Space(n,m)::rest ->
      begin
	let new_trees = join_trees rest in
	  match new_trees with
	    | Chars(p,q)::tail -> Concat([Space(n,m);Chars(p,q)])::tail
	    | Space(p,q)::tail -> Concat([Space(n,m);Space(p,q)])::tail
	    | Concat(l)::tail -> Concat(Space(n,m)::l)::tail 
	    | _ -> Space(n,m)::new_trees
      end
  | Concat(l1)::rest ->
      begin
	let new_trees = join_trees rest in
	  match new_trees with
	    | Chars(p,q)::tail -> Concat(l1@[Chars(p,q)])::tail
	    | Space(p,q)::tail -> Concat(l1@[Space(p,q)])::tail
	    | Concat(l2)::tail -> Concat(l1@l2)::tail 
	    | _ -> Concat(l1)::new_trees
      end
;;

let rec join_leaves = function
(* joins adjacent characters and spaces *)
(* works only on current level, no recursive descent into the nodes *)
  | [] -> []
  | Node(tags,children)::tail ->
      Node(tags,children)::(join_leaves tail)
  | Chars(m,n)::tail -> 
      let jtree = join_leaves tail in
	begin
	  match jtree with
	    | Chars(p,q)::jtail -> Concat([Chars(m,n);Chars(p,q)])::jtail
	    | Space(p,q)::jtail -> Concat([Chars(m,n);Space(p,q)])::jtail
	    | Concat(k)::jtail -> Concat(Chars(m,n)::k)::jtail 
	    | _ -> Chars(m,n)::jtree
	end
  | Space(m,n)::tail -> 
      let jtree = join_leaves tail in
	begin
	  match jtree with
	    | Chars(p,q)::jtail -> Concat([Space(m,n);Chars(p,q)])::jtail
	    | Space(p,q)::jtail -> Concat([Space(m,n); Space(p,q)])::jtail
	    | Concat(k)::jtail -> Concat(Space(m,n)::k)::jtail 
	    | _ -> Space(m,n)::jtree
	end
  | Concat(l)::tail -> 
      let jtree = join_leaves tail in
	begin
	  match jtree with
	    | Chars(p,q)::jtail -> Concat(l@[Chars(p,q)])::jtail
	    | Space(p,q)::jtail -> Concat(l@[Space(p,q)])::jtail
	    | Concat(k)::jtail -> Concat(l@k)::jtail 
	    | _ -> Concat(l)::jtree
	end
;;

(*
let factorize_nodes =
(* factorizes nodes according to likely tags *)
;;
*)

(* bugged *)
let rec redundancy_elim = function
    (* eliminates redundant tags *)
  | (Space _ as tree)
  | (Chars _ as tree) 
  | (Concat _ as tree) -> tree
  | Node(tags,children) ->
      begin
	let new_children = join_leaves children in
	  match new_children with
	    | [child] -> 
		begin
		  let new_child = redundancy_elim child in
		    match new_child with
		      | Node(ntags,childs) -> Node(insert_into_tags ntags tags,childs) (* WARNING *)
		      | _ -> Node(tags,[new_child])
		end
	    | _ -> Node(tags,(*factorize_nodes*) List.map redundancy_elim new_children)
      end
;;

let rec apply_tags context = function
  | [] -> context
  | tag::tail ->
      begin
	match Parser.new_context context tag with
	  | None -> apply_tags context tail
	  | Some ctxt -> apply_tags ctxt tail
      end
;; 

let rec remove_redundancy_tags context = function
  | Node([],children) -> flatten_map (remove_redundancy_tags context) children
  | (Space _ as tree)
  | (Chars _ as tree)
  | (Concat _ as tree) -> [tree]
  | Node(tags,children) ->
      let new_context = apply_tags context tags in
	if context = new_context then
	  flatten_map (remove_redundancy_tags context) (join_leaves children)
	else
	  let new_children = flatten_map (remove_redundancy_tags new_context) (join_leaves children) in
	    begin
	      match new_children with
		| [Space(n,m)] ->
		    if (Parser.new_space_context new_context) = (Parser.new_space_context context) then
		      new_children
		    else
		      [Node(tags,new_children)]
		| _ -> [Node(tags,new_children)]
	    end
;;

let rec print_op_tree = function
  | Node(tags,children) ->
      List.iter (function t -> output_string stderr (Tag.string_of_open_tag t)) tags;
      List.iter print_op_tree children;
      List.iter (function t -> output_string stderr (Tag.string_of_close_tag t)) tags;
      output_string stderr "\n";
      flush stderr;
  | Space(n,m) -> output_string stderr "space"
  | Chars(n,m) -> output_string stderr "chars"
  | Concat(l) -> output_string stderr "concat";List.iter print_op_tree l
;;

let rm_redundancy_tags_trees = flatten_map (remove_redundancy_tags Parser.root_context);;

let slow_naive_join_trees = join_trees

let full_naive_optimization trees = slow_naive_join_trees (rm_redundancy_tags_trees trees);;

let collapse_trees trees canonical size = 
  let opt_trees = List.map op_tree_of_tree (rm_tags_trees trees) in
   (* List.iter print_op_tree opt_trees;flush stderr; *)
    let opt_trees = full_naive_optimization (full_naive_optimization opt_trees) in
     (* List.iter print_op_tree opt_trees;flush stderr; *)
      flatten_map tree_of_op_tree opt_trees
;;












